/* evp_cnf.c */
/*
 * Written by Stephen Henson (steve@openssl.org) for the OpenSSL project
 * 2007.
 */
/* ====================================================================
 * Copyright (c) 2007 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.OpenSSL.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    licensing@OpenSSL.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.OpenSSL.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <openssl/crypto.h>
#include "cryptlib.h"
#include <openssl/conf.h>
#include <openssl/dso.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#ifdef OPENSSL_FIPS
# include <openssl/fips.h>
#endif
#ifdef NETAPP
void alg_module_unload(CONF_MODULE *pmod);
#endif /* NETAPP */
/* Algorithm configuration module. */

static int alg_module_init(CONF_IMODULE *md, const CONF *cnf)
{
#ifndef NETAPP
    int i;
    const char *oid_section;
    STACK_OF(CONF_VALUE) *sktmp;
    CONF_VALUE *oval;
    oid_section = CONF_imodule_get_value(md);
    if (!(sktmp = NCONF_get_section(cnf, oid_section))) {
        EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_ERROR_LOADING_SECTION);
        return 0;
    }
    for (i = 0; i < sk_CONF_VALUE_num(sktmp); i++) {
        oval = sk_CONF_VALUE_value(sktmp, i);
        if (!strcmp(oval->name, "fips_mode")) {
            int m;
            if (!X509V3_get_value_bool(oval, &m)) {
                EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_INVALID_FIPS_MODE);
                return 0;
            }
            if (m > 0) {
#ifdef OPENSSL_FIPS
                if (!FIPS_mode() && !FIPS_mode_set(1)) {
                    EVPerr(EVP_F_ALG_MODULE_INIT,
                           EVP_R_ERROR_SETTING_FIPS_MODE);
                    return 0;
                }
#else
                EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_FIPS_MODE_NOT_SUPPORTED);
                return 0;
#endif
            }
        } else {
            EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_UNKNOWN_OPTION);
            ERR_add_error_data(4, "name=", oval->name,
                               ", value=", oval->value);
        }
    }
    return 1;
#else /* NETAPP */
    int i;
    const char *oid_section;
    STACK_OF(CONF_VALUE) *sktmp;
    CONF_VALUE *oval;
    oid_section = CONF_imodule_get_value(md);
    void *usr_data = NULL;
    ALG_MODULE_DATA  *alg_md_data = NULL;

    /* Set the unload func in parent module */
    CONF_MODULE *pmod = CONF_imodule_get_module(md);
    if(!CONF_module_get_conf_unload_func(pmod))
        CONF_module_set_conf_unload_func(pmod, alg_module_unload);

    if (!(sktmp = NCONF_get_section(cnf, oid_section))) {
        EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_ERROR_LOADING_SECTION);
        return 0;
    }

    CRYPTO_THREADID *cur = OPENSSL_malloc(sizeof(CRYPTO_THREADID));
    CRYPTO_THREADID_current(cur);
    int updated = 0;

    for (i = 0; i < sk_CONF_VALUE_num(sktmp); i++) {
        oval = sk_CONF_VALUE_value(sktmp, i);
        if (!strcmp(oval->name, "fips_mode") ||
                !strcmp(oval->name, "ssl_protocol") ||
                !strcmp(oval->name, "ssl_ciphers")) {

            CRYPTO_w_lock(CRYPTO_LOCK_ALG_CONF);
            usr_data = CONF_module_get_usr_data(pmod);

            if (!usr_data) {
                usr_data = OPENSSL_malloc(sizeof(ALG_MODULE_DATA));
                if (!usr_data)
                    goto err;
                memset(usr_data, 0, sizeof(ALG_MODULE_DATA));

                CONF_module_set_usr_data(pmod, usr_data);
            }

            alg_md_data = (ALG_MODULE_DATA *) usr_data;

            // Set current thread's id
            if(alg_md_data) {
                if (alg_md_data->thr_id) {
                    if (CRYPTO_THREADID_cmp(cur, alg_md_data->thr_id)) {
                        OPENSSL_free(alg_md_data->thr_id);
                        alg_md_data->thr_id = cur;
                        updated = 1;
                    }
                } else {
                    alg_md_data->thr_id = cur;
                    updated = 1;
                }
                if(!alg_md_data->thr_id) {
                    goto err;
                }
            }

            if (!strcmp(oval->name, "fips_mode")) {
                int m;
                if (!X509V3_get_value_bool(oval, &m)) {
                    EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_INVALID_FIPS_MODE);
                    goto err;
                }
                alg_md_data->fips_mode = m;

                /* IMP: We must unlock CRYPTO_LOCK_ALG_CONF before invoking
                   FIPS_mode_set(1) */
                CRYPTO_w_unlock(CRYPTO_LOCK_ALG_CONF);

                if (m > 0) {
#ifdef OPENSSL_FIPS
                    if (!FIPS_module_mode() && !FIPS_mode_set(1)) {
                        EVPerr(EVP_F_ALG_MODULE_INIT,
                            EVP_R_ERROR_SETTING_FIPS_MODE);
                        CRYPTO_w_lock(CRYPTO_LOCK_ALG_CONF);
                        goto err;
                    }
#else
                    EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_FIPS_MODE_NOT_SUPPORTED);
                    CRYPTO_w_lock(CRYPTO_LOCK_ALG_CONF);
                    goto err;
#endif
                }

                /* NOTE: We must continue here to avoid having to unlock the lock */
                continue;

            } /* fips_mode */
            else if (!strcmp(oval->name, "ssl_protocol")) {
                if(alg_md_data->ssl_protocol_list)
                    goto done;
                alg_md_data->ssl_protocol_list = X509V3_parse_list(oval->value);
                if (!alg_md_data->ssl_protocol_list)
                    goto err;
            } /* ssl_protocol */
            else if (!strcmp(oval->name, "ssl_ciphers")) {
                if(alg_md_data->ssl_cipher_list)
                    goto done;
                alg_md_data->ssl_cipher_list = OPENSSL_strdup(oval->value);
                if (!alg_md_data->ssl_cipher_list)
                    goto err;
            } /* ssl_ciphers */

            done:
                CRYPTO_w_unlock(CRYPTO_LOCK_ALG_CONF);
                continue;

            err:
                if (alg_md_data) {
                    if (alg_md_data->thr_id)
                        OPENSSL_free(alg_md_data->thr_id);
                    if (alg_md_data->ssl_protocol_list)
                        sk_CONF_VALUE_pop_free(alg_md_data->ssl_protocol_list, X509V3_conf_free);
                    if (alg_md_data->ssl_cipher_list)
                        OPENSSL_free(alg_md_data->ssl_cipher_list);
                    OPENSSL_free(alg_md_data);
                    CONF_module_set_usr_data(pmod, NULL);
                }
                EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_INSUFFICIENT_MEMORY);
                CRYPTO_w_unlock(CRYPTO_LOCK_ALG_CONF);
                return 0;

        } else {
            EVPerr(EVP_F_ALG_MODULE_INIT, EVP_R_UNKNOWN_OPTION);
            ERR_add_error_data(4, "name=", oval->name,
                               ", value=", oval->value);
        }
    }
    if (!updated) {
        OPENSSL_free(cur);
    }
    return 1;
#endif /* NETAPP */
}
#ifdef NETAPP
void alg_module_unload(CONF_MODULE *pmod)
{
    CRYPTO_w_lock(CRYPTO_LOCK_ALG_CONF);
    void *usr_data = CONF_module_get_usr_data(pmod);
    ALG_MODULE_DATA  *alg_md_data = (ALG_MODULE_DATA *) usr_data;
    if (alg_md_data) {
        if (alg_md_data->thr_id)
            OPENSSL_free(alg_md_data->thr_id);
        if (alg_md_data->ssl_protocol_list)
            sk_CONF_VALUE_pop_free(alg_md_data->ssl_protocol_list, X509V3_conf_free);
        if (alg_md_data->ssl_cipher_list)
            OPENSSL_free(alg_md_data->ssl_cipher_list);
        OPENSSL_free(alg_md_data);
        CONF_module_set_usr_data(pmod, NULL);
    }
    CRYPTO_w_unlock(CRYPTO_LOCK_ALG_CONF);
}

#endif /* NETAPP */

void EVP_add_alg_module(void)
{
    CONF_module_add("alg_section", alg_module_init, 0);
}
